from pathlib import Path

from ._file_filter import FileFilter

__all__ = ("HiddenFilter",)


class HiddenFilter(FileFilter):
    def filter(self, path: Path) -> bool:
        return path.name.startswith(".")



from math import isclose
from unittest.mock import Mock

from baby_steps import given, then, when

from vedro._core import Report, ScenarioResult, VirtualScenario


def make_scenario_result() -> ScenarioResult:
    scenario_ = Mock(VirtualScenario)
    return ScenarioResult(scenario_)


def test_report_defaults():
    with when:
        report = Report()

    with then:
        assert report.total == 0
        assert report.passed == 0
        assert report.failed == 0
        assert report.skipped == 0
        assert report.elapsed == 0.0


def test_report_passed():
    with given:
        report = Report()
        scenario_result = make_scenario_result().mark_passed()

    with when:
        res = report.add_result(scenario_result)

    with then:
        assert res is None
        assert report.total == 1
        assert report.passed == 1
        assert report.failed == 0
        assert report.skipped == 0


def test_report_failed():
    with given:
        report = Report()
        scenario_result = make_scenario_result().mark_failed()

    with when:
        res = report.add_result(scenario_result)

    with then:
        assert res is None
        assert report.total == 1
        assert report.passed == 0
        assert report.failed == 1
        assert report.skipped == 0


def test_report_skipped():
    with given:
        report = Report()
        scenario_result = make_scenario_result().mark_skipped()

    with when:
        res = report.add_result(scenario_result)

    with then:
        assert res is None
        assert report.total == 1
        assert report.passed == 0
        assert report.failed == 0
        assert report.skipped == 1


def test_report_elapsed():
    with given:
        report = Report()
        scenario_result = make_scenario_result()
        scenario_result.set_started_at(1.0)
        scenario_result.set_ended_at(3.0)

    with when:
        res = report.add_result(scenario_result)

    with then:
        assert res is None
        assert isclose(report.elapsed, 2.0)


def test_report_eq():
    with given:
        report1 = Report()
        report2 = Report()

    with when:
        res = report1 == report2

    with then:
        assert res is True


def test_report_eq_with_same_results():
    with given:
        scenario_result1 = make_scenario_result().mark_passed()
        scenario_result2 = make_scenario_result().mark_failed()

        report1 = Report()
        report1.add_result(scenario_result1)
        report1.add_result(scenario_result2)

        report2 = Report()
        report2.add_result(scenario_result1)
        report2.add_result(scenario_result2)

    with when:
        res = report1 == report2

    with then:
        assert res is True


def test_report_eq_with_diff_results():
    with given:
        scenario_result1 = make_scenario_result().mark_passed()
        scenario_result2 = make_scenario_result().mark_failed()
        scenario_result3 = make_scenario_result().mark_failed()

        report1 = Report()
        report1.add_result(scenario_result1)
        report1.add_result(scenario_result2)

        report2 = Report()
        report2.add_result(scenario_result1)
        report2.add_result(scenario_result3)

    with when:
        res = report1 == report2

    with then:
        assert res is False



from ._any_filter import AnyFilter
from ._dunder_filter import DunderFilter
from ._ext_filter import ExtFilter
from ._file_filter import FileFilter
from ._hidden_filter import HiddenFilter

__all__ = ("AnyFilter", "DunderFilter", "ExtFilter", "FileFilter", "HiddenFilter",)



from typing import Any, Type, Union

from ..._scenario import Scenario

__all__ = ("skip",)


def skip(scenario_or_reason: Union[Type[Scenario], str]) -> Any:
    def wrapped(scenario: Type[Scenario]) -> Type[Scenario]:
        setattr(scenario, "__vedro__skipped__", True)
        return scenario

    if isinstance(scenario_or_reason, str):
        return wrapped
    elif issubclass(scenario_or_reason, Scenario):
        return wrapped(scenario_or_reason)
    else:
        raise TypeError()



from ._scenario_assert_rewriter_loader import ScenarioAssertRewriterLoader
from ._scenario_file_loader import ScenarioFileLoader
from ._scenario_loader import ScenarioLoader

__all__ = ("ScenarioLoader", "ScenarioFileLoader", "ScenarioAssertRewriterLoader",)



from asyncio import iscoroutinefunction
from heapq import heappop, heappush
from time import monotonic_ns
from typing import Any, Callable, Dict, List, Tuple, Type

from ..events import Event

__all__ = ("Dispatcher", "Subscriber",)


HandlerType = Callable[..., Any]
RegisteredItemType = Tuple[int, int, HandlerType]


class Subscriber:
    def subscribe(self, dispatcher: "Dispatcher") -> None:
        raise NotImplementedError()


class Dispatcher:
    def __init__(self) -> None:
        self._events: Dict[str, List[RegisteredItemType]] = {}

    def register(self, subscriber: "Subscriber") -> None:
        subscriber.subscribe(self)

    def listen(self, event: Type[Event], handler: HandlerType, priority: int = 0) -> "Dispatcher":
        if event.__name__ not in self._events:
            self._events[event.__name__] = []
        heappush(self._events[event.__name__], (priority, monotonic_ns(), handler))
        return self

    async def fire(self, event: Event) -> None:
        if event.__class__.__name__ not in self._events:
            return
        registered = self._events[event.__class__.__name__]
        registered_copy: List[RegisteredItemType] = []
        while len(registered) > 0:
            priority, registered_at, handler = heappop(registered)
            if iscoroutinefunction(handler):
                await handler(event)
            else:
                handler(event)
            heappush(registered_copy, (priority, registered_at, handler))
        self._events[event.__class__.__name__] = registered_copy



from pathlib import Path
from textwrap import dedent

import pytest
from baby_steps import given, then, when

from vedro._core._scenario_loader import ScenarioFileLoader


@pytest.mark.asyncio
async def test_scenario_file_loader(tmp_path: Path):
    with given:
        path = tmp_path / "scenario.py"
        path.write_text(dedent('''
            import vedro
            class Scenario(vedro.Scenario):
                pass
        '''))

        loader = ScenarioFileLoader()

    with when:
        scenarios = await loader.load(path)

    with then:
        assert len(scenarios) == 1


@pytest.mark.asyncio
async def test_template_scenario_file_loader(tmp_path: Path):
    with given:
        path = tmp_path / "scenario.py"
        path.write_text(dedent('''
            import vedro
            class Scenario(vedro.Scenario):
                @vedro.params("Bob")
                @vedro.params("Alice")
                def __init__(self, user):
                    pass
        '''))

        loader = ScenarioFileLoader()

    with when:
        scenarios = await loader.load(path)

    with then:
        assert len(scenarios) == 2



from ...._core import Dispatcher
from .._reporter import Reporter

__all__ = ("SilentReporter",)


class SilentReporter(Reporter):
    def subscribe(self, dispatcher: Dispatcher) -> None:
        pass



import asyncio
from asyncio import CancelledError
from typing import List, Optional

from ._context import context
from ._core import Dispatcher, Lifecycle, Plugin, Runner, ScenarioDiscoverer
from ._core._scenario_finder import ScenarioFileFinder
from ._core._scenario_finder._file_filters import AnyFilter, DunderFilter, ExtFilter, HiddenFilter
from ._core._scenario_loader import ScenarioAssertRewriterLoader
from ._interface import Interface
from ._params import params
from ._scenario import Scenario
from ._version import version
from .plugins.director import Director, Reporter, RichReporter, SilentReporter
from .plugins.skipper import Skipper, only, skip
from .plugins.terminator import Terminator

__version__ = version
__all__ = ("Scenario", "Interface", "Runner", "run", "only", "skip", "params", "context",)


def run(*, plugins: Optional[List[Plugin]] = None) -> None:
    finder = ScenarioFileFinder(
        file_filter=AnyFilter([
            HiddenFilter(),
            DunderFilter(),
            ExtFilter(only=["py"]),
        ]),
        dir_filter=AnyFilter([
            HiddenFilter(),
            DunderFilter(),
        ])
    )
    loader = ScenarioAssertRewriterLoader()
    discoverer = ScenarioDiscoverer(finder, loader)
    dispatcher = Dispatcher()

    all_plugins = []
    reporters = [
        RichReporter(),
        SilentReporter(),
    ]

    if plugins:
        for plugin in plugins:
            if isinstance(plugin, Reporter):
                reporters.append(plugin)
            else:
                all_plugins.append(plugin)

    all_plugins += [
        Director(reporters),
        Skipper(),
        Terminator(),
    ]

    runner = Runner(dispatcher, (KeyboardInterrupt, SystemExit, CancelledError,))
    lifecycle = Lifecycle(dispatcher, discoverer, runner)
    lifecycle.register_plugins(all_plugins)

    asyncio.run(lifecycle.start())



from pathlib import Path
from typing import List

import pytest

from vedro._core._scenario_finder import ScenarioFileFinder
from vedro._core._scenario_finder._file_filters import DunderFilter, HiddenFilter


def _create_file_tree(tree: List[Path]) -> List[Path]:
    for path in tree:
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_text("")
    return tree


@pytest.mark.asyncio
async def test_scenario_file_finder(tmp_path: Path):
    tree = _create_file_tree([
        (tmp_path / ".DS_Store"),
        (tmp_path / "scenario1.py"),
        (tmp_path / "scenario2.py"),
        (tmp_path / "folder1" / "scenario3.py"),
        (tmp_path / "folder1" / "__pychache__" / "scenario3.pyc"),
        (tmp_path / "folder2" / "__init__.py"),
        (tmp_path / "folder2" / "scenario4.py"),
        (tmp_path / "folder2" / "scenario5.py"),
        (tmp_path / "folder3" / "folder4" / "scenario6.py"),
        (tmp_path / "folder3" / "folder4" / "scenario6.pyc"),
    ])

    finder = ScenarioFileFinder()
    paths = []
    async for path in finder.find(tmp_path):
        paths.append(path)

    assert set(paths) == set(tree)
    assert len(paths) == len(tree)


@pytest.mark.asyncio
async def test_scenario_file_finder_with_file_filter(tmp_path: Path):
    subtree = [
        (tmp_path / "scenario1.py"),
        (tmp_path / "scenario2.py"),
        (tmp_path / ".folder1" / "scenario3.py"),
    ]
    _create_file_tree(subtree + [
        (tmp_path / ".DS_Store"),
        (tmp_path / ".folder1" / ".DS_Store"),
    ])

    finder = ScenarioFileFinder(file_filter=HiddenFilter())
    paths = []
    async for path in finder.find(tmp_path):
        paths.append(path)

    assert set(paths) == set(subtree)
    assert len(paths) == len(subtree)


@pytest.mark.asyncio
async def test_scenario_file_finder_with_dir_filter(tmp_path: Path):
    subtree = [
        (tmp_path / "scenario1.py"),
        (tmp_path / "folder1" / "scenario3.py"),
        (tmp_path / "folder2" / "__init__.py"),
    ]
    _create_file_tree(subtree + [
        (tmp_path / "folder1" / "__pychache__" / "scenario3.pyc"),
    ])

    finder = ScenarioFileFinder(dir_filter=DunderFilter())
    paths = []
    async for path in finder.find(tmp_path):
        paths.append(path)

    assert set(paths) == set(subtree)
    assert len(paths) == len(subtree)



import os
from argparse import ArgumentParser, HelpFormatter
from functools import partial
from pathlib import Path
from typing import List

from ..events import ArgParsedEvent, ArgParseEvent, CleanupEvent, StartupEvent
from ._dispatcher import Dispatcher
from ._plugin import Plugin
from ._report import Report
from ._runner import Runner
from ._scenario_discoverer import ScenarioDiscoverer

__all__ = ("Lifecycle",)


class Lifecycle:
    def __init__(self,
                 dispatcher: Dispatcher,
                 discoverer: ScenarioDiscoverer,
                 runner: Runner) -> None:
        self._dispatcher = dispatcher
        self._discoverer = discoverer
        self._runner = runner
        self._plugins: List[Plugin] = []

    def register_plugins(self, plugins: List[Plugin]) -> None:
        for plugin in plugins:
            self._dispatcher.register(plugin)
            self._plugins.append(plugin)

    async def start(self) -> Report:
        formatter = partial(HelpFormatter, max_help_position=30)
        arg_parser = ArgumentParser("vedro", formatter_class=formatter, add_help=False)

        await self._dispatcher.fire(ArgParseEvent(arg_parser))
        arg_parser.add_argument("-h", "--help",
                                action="help", help="Show this help message and exit")
        args = arg_parser.parse_args()
        await self._dispatcher.fire(ArgParsedEvent(args))

        start_dir = os.path.relpath(Path("scenarios"))
        scenarios = await self._discoverer.discover(Path(start_dir))
        await self._dispatcher.fire(StartupEvent(scenarios))

        report = await self._runner.run(scenarios)

        await self._dispatcher.fire(CleanupEvent(report))

        return report

    def __repr__(self) -> str:
        cls_name = self.__class__.__name__
        return f"{cls_name}({self._dispatcher!r}, {self._discoverer!r}, {self._runner!r})"



from pathlib import Path
from typing import List, Optional

from ._file_filter import FileFilter

__all__ = ("ExtFilter",)


class ExtFilter(FileFilter):
    def __init__(self, *,
                 only: Optional[List[str]] = None,
                 ignore: Optional[List[str]] = None) -> None:
        self._only = only or None
        self._ignore = ignore or None
        if (self._only is not None) and (self._ignore is not None):
            raise ValueError("Use 'only' or 'ignore' (not both)")

    def filter(self, path: Path) -> bool:
        if self._only:
            for suffix in self._only:
                if path.name.endswith(suffix):
                    return False
            return True
        if self._ignore:
            for suffix in self._ignore:
                if path.name.endswith(suffix):
                    return True
            return False
        return False



from argparse import ArgumentParser, Namespace
from unittest.mock import Mock, call

import pytest
from baby_steps import given, then, when

from vedro._core import Dispatcher
from vedro.events import ArgParseEvent
from vedro.plugins.director import Director, Reporter


@pytest.fixture()
def dispatcher():
    return Dispatcher()


@pytest.mark.asyncio
async def test_director_plugin(*, dispatcher: Dispatcher):
    with given:
        director = Director()
        director.subscribe(dispatcher)
        event = ArgParseEvent(ArgumentParser())

    with when:
        res = await dispatcher.fire(event)

    with then:
        assert res is None


@pytest.mark.asyncio
async def test_director_plugin_with_default_reporter(*, dispatcher: Dispatcher):
    with given:
        reporter_ = Mock(Reporter)
        reporter_.name = "reporter"
        director = Director([reporter_])
        director.subscribe(dispatcher)
        event = ArgParseEvent(ArgumentParser())

    with when:
        res = await dispatcher.fire(event)

    with then:
        assert res is None
        assert reporter_.mock_calls == [call.subscribe(dispatcher)]


@pytest.mark.asyncio
async def test_director_plugin_with_reporter(*, dispatcher: Dispatcher):
    with given:
        reporter1_, reporter2_ = Mock(Reporter), Mock(Reporter)
        reporter1_.name = "reporter1"
        reporter2_.name = "reporter2"
        director = Director([reporter1_, reporter2_])
        director.subscribe(dispatcher)

        args = Namespace(reporters=[reporter2_.name])
        arg_parser = Mock(ArgumentParser, parse_known_args=Mock(return_value=(args, [])))
        event = ArgParseEvent(arg_parser)

    with when:
        res = await dispatcher.fire(event)

    with then:
        assert res is None
        assert reporter1_.mock_calls == []
        assert reporter2_.mock_calls == [call.subscribe(dispatcher)]


@pytest.mark.asyncio
async def test_director_plugin_with_reporters(*, dispatcher: Dispatcher):
    with given:
        reporter1_, reporter2_ = Mock(Reporter), Mock(Reporter)
        reporter1_.name = "reporter1"
        reporter2_.name = "reporter2"
        director = Director([reporter1_, reporter2_])
        director.subscribe(dispatcher)

        args = Namespace(reporters=[reporter2_.name, reporter1_.name])
        arg_parser = Mock(ArgumentParser, parse_known_args=Mock(return_value=(args, [])))
        event = ArgParseEvent(arg_parser)

    with when:
        res = await dispatcher.fire(event)

    with then:
        assert res is None
        assert reporter1_.mock_calls == [call.subscribe(dispatcher)]
        assert reporter2_.mock_calls == [call.subscribe(dispatcher)]



import os
from typing import List, Union

from ..._core import Dispatcher, Plugin, VirtualScenario
from ...events import ArgParsedEvent, ArgParseEvent, StartupEvent

__all__ = ("Skipper",)


class Skipper(Plugin):
    def __init__(self) -> None:
        self._subject: Union[str, None] = None
        self._specified: List[str] = []
        self._ignored: List[str] = []

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(StartupEvent, self.on_startup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        event.arg_parser.add_argument("file_or_dir", nargs='*', default=["."],
                                      help="Select scenarios in a given file or directory")
        event.arg_parser.add_argument("-i", "--ignore", nargs='+', default=[],
                                      help="Skip scenarios in a given file or directory")
        event.arg_parser.add_argument("--subject", help="Select scenarios with a given subject")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._subject = event.args.subject

        for file_or_dir in event.args.file_or_dir:
            path = self._normalize_path(file_or_dir)
            assert os.path.isdir(path) or os.path.isfile(path), f"{path!r} does not exist"
            self._specified.append(path)

        for file_or_dir in event.args.ignore:
            path = self._normalize_path(file_or_dir)
            assert os.path.isdir(path) or os.path.isfile(path), f"{path!r} does not exist"
            self._ignored.append(path)

    def _normalize_path(self, file_or_dir: str) -> str:
        path = os.path.normpath(file_or_dir)
        if os.path.isabs(path):
            return path
        if os.path.commonpath(["scenarios", path]) != "scenarios":
            path = os.path.join("scenarios", path)
        return os.path.abspath(path)

    def _is_scenario_skipped(self, scenario: VirtualScenario) -> bool:
        if getattr(scenario._orig_scenario, "__vedro__skipped__", False):
            return True
        template = getattr(scenario._orig_scenario, "__vedro__template__", None)
        if getattr(template, "__vedro__skipped__", False):
            return True
        if self._subject and scenario.subject != self._subject:
            return True
        if not self._is_scenario_specified(scenario):
            return True
        if self._is_scenario_ignored(scenario):
            return True
        return False

    def _is_scenario_special(self, scenario: VirtualScenario) -> bool:
        if getattr(scenario._orig_scenario, "__vedro__only__", False):
            return True
        template = getattr(scenario._orig_scenario, "__vedro__template__", None)
        if getattr(template, "__vedro__only__", False):
            return True
        return False

    def _is_scenario_specified(self, scenario: VirtualScenario) -> bool:
        for path in self._specified:
            if os.path.commonpath([path, scenario.path]) == path:
                return True
        return False

    def _is_scenario_ignored(self, scenario: VirtualScenario) -> bool:
        for path in self._ignored:
            if os.path.commonpath([path, scenario.path]) == path:
                return True
        return False

    def on_startup(self, event: StartupEvent) -> None:
        special_scenarios = set()
        for scenario in event.scenarios:
            if self._is_scenario_skipped(scenario):
                scenario.skip()
            elif self._is_scenario_special(scenario):
                special_scenarios.add(scenario.unique_id)

        if len(special_scenarios) > 0:
            for scenario in event.scenarios:
                if scenario.unique_id not in special_scenarios:
                    scenario.skip()



import pytest
from baby_steps import given, then, when

from vedro.plugins.director import Reporter


@pytest.mark.parametrize(("cls_name", "reporter_name"), [
    ("WordReporter", "word"),
    ("TwoWordsReporter", "two_words"),
    ("ReportGenerator", "report_generator"),
])
def test_reporter_name(cls_name: str, reporter_name: str):
    with given:
        cls = type(cls_name, (Reporter,), {})

    with when:
        reporter = cls()

    with then:
        assert reporter.name == reporter_name



import sys

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock
else:
    from asynctest.mock import CoroutineMock as AsyncMock

from argparse import Namespace
from pathlib import Path
from unittest.mock import Mock, call, patch

import pytest
from baby_steps import given, then, when

from vedro._core import Dispatcher, Lifecycle, Plugin, Report, Runner, ScenarioDiscoverer
from vedro.events import ArgParsedEvent, CleanupEvent, StartupEvent


@pytest.fixture()
def dispatcher_():
    return AsyncMock(Dispatcher)


@pytest.fixture()
def discoverer_():
    discoverer = Mock(ScenarioDiscoverer)
    return discoverer


@pytest.fixture()
def runner_(dispatcher_: Dispatcher):
    report = Report()
    return Mock(Runner(dispatcher_), run=AsyncMock(return_value=report))


def test_lifecycle_register_plugins(*, dispatcher_: Dispatcher,
                                    discoverer_: ScenarioDiscoverer, runner_: Runner):
    with given:
        lifecycle = Lifecycle(dispatcher_, discoverer_, runner_)
        plugin1 = Mock(Plugin)
        plugin2 = Mock(Plugin)

    with when:
        res = lifecycle.register_plugins([plugin1, plugin2])

    with then:
        assert res is None
        assert dispatcher_.mock_calls == [
            call.register(plugin1),
            call.register(plugin2),
        ]


@pytest.mark.asyncio
async def test_lifecycle_register_start(*, dispatcher_: Dispatcher,
                                        discoverer_: ScenarioDiscoverer, runner_: Runner):
    with given:
        scenarios = []
        discoverer_.discover = AsyncMock(return_value=scenarios)

        lifecycle = Lifecycle(dispatcher_, discoverer_, runner_)
        namespace = Namespace()

    with when, patch("argparse.ArgumentParser.parse_args", return_value=namespace):
        report = await lifecycle.start()

    with then:
        assert isinstance(report, Report)
        assert discoverer_.mock_calls == [
            call.discover(Path("scenarios"))
        ]
        assert runner_.mock_calls == [
            call.run(scenarios),
        ]
        assert dispatcher_.mock_calls[1:] == [
            call.fire(ArgParsedEvent(namespace)),
            call.fire(StartupEvent(scenarios)),
            call.fire(CleanupEvent(report)),
        ]


def test_lifecycle_repr(*, dispatcher_: Dispatcher,
                        discoverer_: ScenarioDiscoverer, runner_: Runner):
    with when:
        lifecycle = Lifecycle(dispatcher_, discoverer_, runner_)

    with then:
        assert repr(lifecycle) == f"Lifecycle({dispatcher_!r}, {discoverer_!r}, {runner_!r})"



import sys

import pytest

if sys.version_info >= (3, 8):
    from unittest.mock import AsyncMock
else:
    from asynctest.mock import CoroutineMock as AsyncMock

from types import MethodType
from unittest.mock import Mock, call, sentinel

from baby_steps import given, then, when

from vedro._core import VirtualStep


def test_virtual_step_name():
    with given:
        method_ = Mock(MethodType)
        method_.__name__ = "<name>"
        step = VirtualStep(method_)

    with when:
        res = step.name

    with then:
        assert res == method_.__name__


def test_virtual_step_is_coro_sync():
    with given:
        method_ = Mock(MethodType, __code__=Mock(co_flags=0))  # python 3.7 specific
        step = VirtualStep(method_)

    with when:
        res = step.is_coro()

    with then:
        assert res is False


def test_virtual_step_is_coro_async():
    with given:
        method_ = AsyncMock(MethodType)
        step = VirtualStep(method_)

    with when:
        res = step.is_coro()

    with then:
        assert res is True


def test_virtual_step_call_sync():
    with given:
        method_ = Mock(MethodType, return_value=sentinel)
        step = VirtualStep(method_)

    with when:
        res = step()

    with then:
        assert res == sentinel
        assert method_.mock_calls == [call()]


@pytest.mark.asyncio
async def test_virtual_step_call_async():
    with given:
        method_ = AsyncMock(MethodType, return_value=sentinel)
        step = VirtualStep(method_)

    with when:
        res = await step()

    with then:
        assert res == sentinel
        assert method_.mock_calls == [call()]


def test_virtual_step_repr():
    with given:
        method_ = Mock(MethodType)
        step = VirtualStep(method_)

    with when:
        res = repr(step)

    with then:
        assert res == f"VirtualStep({method_!r})"


def test_virtual_step_eq():
    with given:
        method_ = Mock(MethodType)
        step1 = VirtualStep(method_)
        step2 = VirtualStep(method_)

    with when:
        res = step1 == step2

    with then:
        assert res is True


def test_virtual_step_not_eq():
    with given:
        method1_ = Mock(MethodType)
        step1 = VirtualStep(method1_)
        method2_ = Mock(MethodType)
        step2 = VirtualStep(method2_)

    with when:
        res = step1 == step2

    with then:
        assert res is False



from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Type

from ..._scenario import Scenario

__all__ = ("ScenarioLoader",)


class ScenarioLoader(ABC):
    @abstractmethod
    async def load(self, path: Path) -> List[Type[Scenario]]:
        pass



from pathlib import Path

import pytest
from pytest import raises

from vedro._core._scenario_finder._file_filters import ExtFilter


@pytest.mark.parametrize(("value", "expected"), [
    ("/tmp", False),
    ("/tmp/main.py", False),
    ("init.sh", False),
])
def test_ext_file_filter(value: str, expected: bool):
    ext_filter = ExtFilter()

    filtered = ext_filter.filter(Path(value))
    assert filtered is expected


@pytest.mark.parametrize(("value", "expected"), [
    ("main.py", False),
    ("init.sh", False),
    ("/tmp/main.py", False),
    #
    ("main.pyc", True),
    ("/tmp/main.pyc", True),
    ("/tmp", True),
])
def test_ext_file_filter_only(value: str, expected: bool):
    ext_filter = ExtFilter(only=[".py", ".sh"])

    filtered = ext_filter.filter(Path(value))
    assert filtered is expected


@pytest.mark.parametrize(("value", "expected"), [
    ("main.py", True),
    ("init.sh", True),
    ("/tmp/main.py", True),
    #
    ("main.pyc", False),
    ("/tmp/main.pyc", False),
    ("/tmp", False),
])
def test_ext_file_filter_ignore(value: str, expected: bool):
    ext_filter = ExtFilter(ignore=[".py", ".sh"])

    filtered = ext_filter.filter(Path(value))
    assert filtered is expected


def test_ext_file_filter_only_and_ignore():
    with raises(Exception) as exc_info:
        ExtFilter(only=[".py"], ignore=[".sh"])

    assert exc_info.type is ValueError
    assert str(exc_info.value) == "Use 'only' or 'ignore' (not both)"



from pathlib import Path
from typing import List
from unittest.mock import Mock

import pytest

from vedro._core._scenario_finder._file_filters import AnyFilter, FileFilter


def true_filter():
    return Mock(FileFilter, filter=Mock(return_value=True))


def false_filter():
    return Mock(FileFilter, filter=Mock(return_value=False))


@pytest.mark.parametrize(("filters", "expected"), [
    ([], False),
    ([true_filter()], True),
    ([false_filter()], False),
    ([true_filter(), true_filter()], True),
    ([true_filter(), false_filter()], True),
    ([false_filter(), true_filter()], True),
    ([false_filter(), false_filter()], False),
])
def test_and_file_filter(filters: List[FileFilter], expected: bool):
    file_filter = AnyFilter(filters)
    filtered = file_filter.filter(Path("/tmp"))
    assert filtered is expected



import json
import os
from traceback import format_exception
from types import TracebackType
from typing import Any, Callable, Dict, Generator, Tuple, Union

from rich.console import Console
from rich.style import Style
from rich.traceback import Traceback

import vedro

from ...._core import Dispatcher, ScenarioResult, StepResult
from ....events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioRunEvent,
    ScenarioSkippedEvent,
    StartupEvent,
)
from .._reporter import Reporter
from .utils import make_console

__all__ = ("RichReporter",)


class RichReporter(Reporter):
    def __init__(self, console_factory: Callable[[], Console] = make_console) -> None:
        self._console = console_factory()
        self._verbosity = 0
        self._tb_show_internal_calls = False
        self._tb_show_locals = False
        self._show_timings = False
        self._namespace: Union[str, None] = None

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(StartupEvent, self.on_startup) \
                  .listen(ScenarioRunEvent, self.on_scenario_run) \
                  .listen(ScenarioSkippedEvent, self.on_scenario_skipped) \
                  .listen(ScenarioPassedEvent, self.on_scenario_passed) \
                  .listen(ScenarioFailedEvent, self.on_scenario_failed) \
                  .listen(CleanupEvent, self.on_cleanup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        help_message = ("Increase verbosity level "
                        "(-v: show steps, -vv: show exception, -vvv: show scope)")
        event.arg_parser.add_argument("-v", "--verbose",
                                      action="count",
                                      default=self._verbosity,
                                      help=help_message)
        event.arg_parser.add_argument("--show-timings",
                                      action='store_true',
                                      default=False,
                                      help="Show the elapsed time of each scenario")
        event.arg_parser.add_argument("--tb-show-internal-calls",
                                      action='store_true',
                                      default=False,
                                      help="Show internal calls in the traceback output")
        event.arg_parser.add_argument("--tb-show-locals",
                                      action='store_true',
                                      default=False,
                                      help="Show local variables in the traceback output")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._verbosity = event.args.verbose
        self._show_timings = event.args.show_timings
        self._tb_show_internal_calls = event.args.tb_show_internal_calls
        self._tb_show_locals = event.args.tb_show_locals

    def on_startup(self, event: StartupEvent) -> None:
        self._console.out("Scenarios")

    def on_scenario_skipped(self, event: ScenarioSkippedEvent) -> None:
        pass

    def on_scenario_run(self, event: ScenarioRunEvent) -> None:
        if event.scenario_result.scenario_namespace != self._namespace:
            self._namespace = event.scenario_result.scenario_namespace
            self._console.out(f"* {self._namespace}", style=Style(bold=True))

    def _print_scenario_subject(self, scenario_result: ScenarioResult) -> None:
        if scenario_result.is_passed():
            subject = f" ✔ {scenario_result.scenario_subject}"
            style = Style(color="green")
        elif scenario_result.is_failed():
            subject = f" ✗ {scenario_result.scenario_subject}"
            style = Style(color="red")
        else:
            return

        if self._show_timings:
            if (scenario_result.started_at is None) or (scenario_result.ended_at is None):
                elapsed = 0.0
            else:
                elapsed = scenario_result.ended_at - scenario_result.started_at
            self._console.out(subject, style=style, end="")
            self._console.out(f" ({elapsed:.2f}s)", style=Style(color="grey50"))
        else:
            self._console.out(subject, style=style)

    def _print_step_name(self, step_result: StepResult) -> None:
        if step_result.is_passed():
            name = f"    ✔ {step_result.step_name}"
            style = Style(color="green")
        elif step_result.is_failed():
            name = f"    ✗ {step_result.step_name}"
            style = Style(color="red")
        else:
            return

        if self._show_timings:
            if (step_result.started_at is None) or (step_result.ended_at is None):
                elapsed = 0.0
            else:
                elapsed = step_result.ended_at - step_result.started_at
            self._console.out(name, style=style, end="")
            self._console.out(f" ({elapsed:.2f}s)", style=Style(color="grey50"))
        else:
            self._console.out(name, style=style)

    def on_scenario_passed(self, event: ScenarioPassedEvent) -> None:
        self._print_scenario_subject(event.scenario_result)

    def _filter_locals(self, local_variables: Dict[str, Any]) -> Dict[str, Any]:
        filtered_locals = {}
        for key, val in local_variables.items():
            # self is always "Scenario"
            if key == "self":
                continue
            # assertion rewriter stuff
            if key.startswith("@py"):
                continue
            filtered_locals[key] = val
        return filtered_locals

    def _print_exception(self, exception: BaseException, traceback: TracebackType) -> None:
        if not self._tb_show_internal_calls:
            root = os.path.dirname(vedro.__file__)
            while traceback.tb_next is not None:
                filename = os.path.abspath(traceback.tb_frame.f_code.co_filename)
                if os.path.commonpath([root, filename]) != root:
                    break
                traceback = traceback.tb_next

        if self._tb_show_locals:
            trace = Traceback.extract(type(exception), exception, traceback, show_locals=True)
            for stack in trace.stacks:
                for frame in stack.frames:
                    if frame.locals:
                        frame.locals = self._filter_locals(frame.locals)
            self._console.print(Traceback(trace))
            self._console.out(" ")
        else:
            formatted = format_exception(type(exception), exception, traceback)
            self._console.print("".join(formatted), style=Style(color="yellow"))

    def _format_scope(self, scope: Dict[Any, Any]) -> Generator[Tuple[str, str], None, None]:
        for key, val in scope.items():
            try:
                val_repr = json.dumps(val, ensure_ascii=False, indent=4)
            except:  # noqa: E722
                val_repr = repr(val)
            yield str(key), val_repr

    def _print_scope(self, scope: Dict[Any, Any]) -> None:
        self._console.out("Scope:", style=Style(color="blue", bold=True))
        for key, val in self._format_scope(scope):
            self._console.out(f" {key}: ", end="", style=Style(color="blue"))
            self._console.out(val)

    def on_scenario_failed(self, event: ScenarioFailedEvent) -> None:
        self._print_scenario_subject(event.scenario_result)

        if self._verbosity > 0:
            for step_result in event.scenario_result.step_results:
                self._print_step_name(step_result)

        if self._verbosity > 1:
            for step_result in event.scenario_result.step_results:
                if step_result.exc_info:
                    self._print_exception(step_result.exc_info.value,
                                          step_result.exc_info.traceback)

        if self._verbosity > 2:
            if event.scenario_result.scope:
                self._print_scope(event.scenario_result.scope)
                self._console.out(" ")

    def on_cleanup(self, event: CleanupEvent) -> None:
        if event.report.failed == 0 and event.report.passed > 0:
            style = Style(color="green", bold=True)
        else:
            style = Style(color="red", bold=True)
        self._console.out(" ")
        self._console.out(f"# {event.report.total} scenarios, "
                          f"{event.report.passed} passed, "
                          f"{event.report.failed} failed, "
                          f"{event.report.skipped} skipped",
                          style=style,
                          end="")
        self._console.out(f" ({event.report.elapsed:.2f}s)", style=Style(color="blue"))
